;+LIBPKIXprivate {
;+    global:
;+# libpkix functions
;+# May not become part of the NSS public API. Needed for unit testing for now.
NSS_Get_PKIX_PL_LDAPMessageTemplate;
PKIX_PL_LDAPMessageTemplate;
PKIX_ALLOC_ERROR;
PKIX_BuildChain;
pkix_BuildResult_Create;
PKIX_BuildResult_GetCertChain;
PKIX_BuildResult_GetValidateResult;
PKIX_CertChainChecker_Create;
PKIX_CertChainChecker_GetCertChainCheckerState;
PKIX_CertChainChecker_GetCheckCallback;
PKIX_CertChainChecker_GetSupportedExtensions;
PKIX_CertChainChecker_IsForwardCheckingSupported;
PKIX_CertChainChecker_IsForwardDirectionExpected;
PKIX_CertChainChecker_SetCertChainCheckerState;
PKIX_CertSelector_Create;
PKIX_CertSelector_GetCertSelectorContext;
PKIX_CertSelector_GetCommonCertSelectorParams;
PKIX_CertSelector_GetMatchCallback;
PKIX_CertSelector_SetCommonCertSelectorParams;
PKIX_CertStore_CertContinue;
PKIX_CertStore_Create;
PKIX_CertStore_CrlContinue;
PKIX_CertStore_GetCertCallback;
PKIX_CertStore_GetCertStoreContext;
PKIX_CertStore_GetCRLCallback;
PKIX_CertStore_GetTrustCallback;
pkix_CheckType;
PKIX_ComCertSelParams_AddPathToName;
PKIX_ComCertSelParams_AddSubjAltName;
PKIX_ComCertSelParams_Create;
PKIX_ComCertSelParams_GetAuthorityKeyIdentifier;
PKIX_ComCertSelParams_GetBasicConstraints;
PKIX_ComCertSelParams_GetCertificate;
PKIX_ComCertSelParams_GetCertificateValid;
PKIX_ComCertSelParams_GetExtendedKeyUsage;
PKIX_ComCertSelParams_GetIssuer;
PKIX_ComCertSelParams_GetKeyUsage;
PKIX_ComCertSelParams_GetMatchAllSubjAltNames;
PKIX_ComCertSelParams_GetNameConstraints;
PKIX_ComCertSelParams_GetPathToNames;
PKIX_ComCertSelParams_GetPolicy;
PKIX_ComCertSelParams_GetSerialNumber;
PKIX_ComCertSelParams_GetSubjAltNames;
PKIX_ComCertSelParams_GetSubject;
PKIX_ComCertSelParams_GetSubjKeyIdentifier;
PKIX_ComCertSelParams_GetSubjPKAlgId;
PKIX_ComCertSelParams_GetSubjPubKey;
PKIX_ComCertSelParams_GetVersion;
PKIX_ComCertSelParams_SetAuthorityKeyIdentifier;
PKIX_ComCertSelParams_SetBasicConstraints;
PKIX_ComCertSelParams_SetCertificate;
PKIX_ComCertSelParams_SetCertificateValid;
PKIX_ComCertSelParams_SetExtendedKeyUsage;
PKIX_ComCertSelParams_SetIssuer;
PKIX_ComCertSelParams_SetKeyUsage;
PKIX_ComCertSelParams_SetMatchAllSubjAltNames;
PKIX_ComCertSelParams_SetNameConstraints;
PKIX_ComCertSelParams_SetPathToNames;
PKIX_ComCertSelParams_SetPolicy;
PKIX_ComCertSelParams_SetSerialNumber;
PKIX_ComCertSelParams_SetSubjAltNames;
PKIX_ComCertSelParams_SetSubject;
PKIX_ComCertSelParams_SetSubjKeyIdentifier;
PKIX_ComCertSelParams_SetSubjPKAlgId;
PKIX_ComCertSelParams_SetSubjPubKey;
PKIX_ComCertSelParams_SetVersion;
PKIX_ComCRLSelParams_AddIssuerName;
PKIX_ComCRLSelParams_Create;
PKIX_ComCRLSelParams_GetCertificateChecking;
PKIX_ComCRLSelParams_GetDateAndTime;
PKIX_ComCRLSelParams_GetIssuerNames;
PKIX_ComCRLSelParams_GetMaxCRLNumber;
PKIX_ComCRLSelParams_GetMinCRLNumber;
PKIX_ComCRLSelParams_SetCertificateChecking;
PKIX_ComCRLSelParams_SetDateAndTime;
PKIX_ComCRLSelParams_SetIssuerNames;
PKIX_ComCRLSelParams_SetMaxCRLNumber;
PKIX_ComCRLSelParams_SetMinCRLNumber;
PKIX_CRLSelector_Create;
PKIX_CRLSelector_GetCommonCRLSelectorParams;
PKIX_CRLSelector_GetCRLSelectorContext;
PKIX_CRLSelector_GetMatchCallback;
PKIX_CRLSelector_SetCommonCRLSelectorParams;
pkix_DefaultRevChecker_Initialize;
PKIX_Error_Create;
PKIX_Error_GetCause;
PKIX_Error_GetDescription;
PKIX_Error_GetErrorClass;
PKIX_Error_GetSupplementaryInfo;
PKIX_ERRORCLASSNAMES DATA ;
PKIX_Initialize;
PKIX_Initialize_SetConfigDir;
PKIX_List_AppendItem;
PKIX_List_Create;
PKIX_List_DeleteItem;
PKIX_List_GetItem;
PKIX_List_GetLength;
PKIX_List_InsertItem;
PKIX_List_IsEmpty;
PKIX_List_IsImmutable;
PKIX_List_ReverseList;
PKIX_List_SetImmutable;
PKIX_List_SetItem;
PKIX_AddLogger;
PKIX_GetLoggers;
PKIX_SetLoggers;
PKIX_Logger_Create;
PKIX_Logger_GetLoggerContext;
PKIX_Logger_GetLoggingComponent;
PKIX_Logger_GetLogCallback;
PKIX_Logger_GetMaxLoggingLevel;
PKIX_Logger_SetLoggingComponent;
PKIX_Logger_SetMaxLoggingLevel;
PKIX_OcspChecker_Initialize;
PKIX_OcspChecker_SetOCSPResponder;
PKIX_OcspChecker_SetPasswordInfo;
PKIX_OcspChecker_SetVerifyFcn;
PKIX_PL_AcquireReaderLock;
PKIX_PL_AcquireWriterLock;
PKIX_PL_BasicConstraints_GetCAFlag;
PKIX_PL_BasicConstraints_GetPathLenConstraint;
PKIX_PL_BigInt_Create;
PKIX_PL_ByteArray_Create;
PKIX_PL_ByteArray_GetLength;
PKIX_PL_ByteArray_GetPointer;
PKIX_PL_Cert_AreCertPoliciesCritical;
PKIX_PL_Cert_CheckNameConstraints;
PKIX_PL_Cert_CheckValidity;
PKIX_PL_Cert_Create;
pkix_pl_Cert_CreateToList;
pkix_pl_Cert_CreateWithNSSCert;
PKIX_PL_Cert_GetAuthorityInfoAccess;
PKIX_PL_Cert_GetAuthorityKeyIdentifier;
PKIX_PL_Cert_GetBasicConstraints;
PKIX_PL_Cert_GetCriticalExtensionOIDs;
PKIX_PL_Cert_GetExtendedKeyUsage;
PKIX_PL_Cert_GetInhibitAnyPolicy;
PKIX_PL_Cert_GetIssuer;
PKIX_PL_Cert_GetNameConstraints;
PKIX_PL_Cert_GetPolicyInformation;
PKIX_PL_Cert_GetPolicyMappingInhibited;
PKIX_PL_Cert_GetPolicyMappings;
PKIX_PL_Cert_GetRequireExplicitPolicy;
PKIX_PL_Cert_GetSerialNumber;
PKIX_PL_Cert_GetSubject;
PKIX_PL_Cert_GetSubjectAltNames;
PKIX_PL_Cert_GetSubjectInfoAccess;
PKIX_PL_Cert_GetSubjectKeyIdentifier;
PKIX_PL_Cert_GetSubjectPublicKey;
PKIX_PL_Cert_GetSubjectPublicKeyAlgId;
PKIX_PL_Cert_GetVersion;
PKIX_PL_Cert_IsCertTrusted;
PKIX_PL_Cert_MergeNameConstraints;
PKIX_PL_Cert_VerifyKeyUsage;
PKIX_PL_Cert_VerifySignature;
PKIX_PL_CertPolicyInfo_GetPolicyId;
PKIX_PL_CertPolicyInfo_GetPolQualifiers;
PKIX_PL_CertPolicyMap_GetIssuerDomainPolicy;
PKIX_PL_CertPolicyMap_GetSubjectDomainPolicy;
PKIX_PL_CollectionCertStore_Create;
PKIX_PL_CRL_Create;
pkix_pl_CRL_CreateToList;
PKIX_PL_CRL_GetCriticalExtensionOIDs;
PKIX_PL_CRL_GetCRLEntryForSerialNumber;
PKIX_PL_CRL_GetIssuer;
PKIX_PL_CRL_VerifySignature;
PKIX_PL_CRLEntry_GetCriticalExtensionOIDs;
PKIX_PL_CRLEntry_GetCRLEntryReasonCode;
PKIX_PL_Date_Create_UTCTime;
pkix_pl_Date_CreateFromPRTime;
pkix_pl_Date_GetPRTime;
PKIX_PL_EkuChecker_Initialize;
PKIX_PL_Free;
PKIX_PL_GeneralName_Create;
PKIX_PL_GetString;
PKIX_PL_HashTable_Add;
PKIX_PL_HashTable_Create;
PKIX_PL_HashTable_Lookup;
PKIX_PL_HashTable_Remove;
PKIX_PL_HttpCertStore_Create;
pkix_pl_HttpCertStore_CreateWithAsciiName;
PKIX_PL_Initialize;
PKIX_PL_InfoAccess_GetMethod;
PKIX_PL_InfoAccess_GetLocation;
PKIX_PL_InfoAccess_GetLocationType;
PKIX_PL_LdapCertStore_Create;
PKIX_PL_LdapClient_InitiateRequest;
PKIX_PL_LdapClient_ResumeRequest;
PKIX_PL_LdapDefaultClient_AbandonRequest;
PKIX_PL_LdapDefaultClient_Create;
PKIX_PL_LdapDefaultClient_CreateByName;
pkix_pl_LdapResponse_Create;
PKIX_PL_Malloc;
PKIX_PL_Memcpy;
PKIX_PL_MonitorLock_Create;
PKIX_PL_MonitorLock_Enter;
PKIX_PL_MonitorLock_Exit;
PKIX_PL_Mutex_Create;
PKIX_PL_Mutex_Lock;
PKIX_PL_Mutex_Unlock;
PKIX_PL_NssContext_Create;
PKIX_PL_Object_Alloc;
PKIX_PL_Object_Compare;
PKIX_PL_Object_DecRef;
PKIX_PL_Object_Duplicate;
PKIX_PL_Object_Equals;
PKIX_PL_Object_GetType;
PKIX_PL_Object_Hashcode;
PKIX_PL_Object_IncRef;
PKIX_PL_Object_IsTypeRegistered;
PKIX_PL_Object_Lock;
PKIX_PL_Object_RegisterType;
PKIX_PL_Object_ToString;
PKIX_PL_Object_Unlock;
PKIX_PL_OID_Create;
pkix_pl_OcspRequest_Create;
pkix_pl_OcspResponse_Create;
pkix_pl_OcspResponse_Decode;
pkix_pl_OcspResponse_GetStatus;
pkix_pl_OcspResponse_GetStatusForCert;
PKIX_PL_OcspResponse_UseBuildChain;
pkix_pl_OcspResponse_VerifySignature;
PKIX_PL_Pk11CertStore_Create;
PKIX_PL_PolicyQualifier_GetPolicyQualifierId;
PKIX_PL_PolicyQualifier_GetQualifier;
PKIX_PL_PublicKey_MakeInheritedDSAPublicKey;
PKIX_PL_PublicKey_NeedsDSAParameters;
PKIX_PL_Realloc;
PKIX_PL_ReleaseReaderLock;
PKIX_PL_ReleaseWriterLock;
PKIX_PL_RWLock_Create;
PKIX_PL_Shutdown;
pkix_pl_Socket_Create;
pkix_pl_Socket_CreateByName;
pkix_pl_Socket_GetCallbackList;
PKIX_PL_Sprintf;
PKIX_PL_String_Create;
PKIX_PL_String_GetEncoded;
PKIX_PL_X500Name_Create;
pkix_pl_X500Name_GetCommonName;
pkix_pl_X500Name_GetCountryName;
pkix_pl_X500Name_GetOrgName;
PKIX_PL_X500Name_Match;
pkix_PolicyNode_AddToParent;
pkix_PolicyNode_Create;
PKIX_PolicyNode_GetChildren;
PKIX_PolicyNode_GetDepth;
PKIX_PolicyNode_GetExpectedPolicies;
PKIX_PolicyNode_GetParent;
PKIX_PolicyNode_GetPolicyQualifiers;
PKIX_PolicyNode_GetValidPolicy;
PKIX_PolicyNode_IsCritical;
pkix_PolicyNode_Prune;
PKIX_ProcessingParams_AddCertChainChecker;
PKIX_ProcessingParams_AddCertStore;
PKIX_ProcessingParams_AddRevocationChecker;
PKIX_ProcessingParams_Create;
PKIX_ProcessingParams_GetCertChainCheckers;
PKIX_ProcessingParams_GetCertStores;
PKIX_ProcessingParams_GetDate;
PKIX_ProcessingParams_GetHintCerts;
PKIX_ProcessingParams_GetInitialPolicies;
PKIX_ProcessingParams_GetPolicyQualifiersRejected;
PKIX_ProcessingParams_GetResourceLimits;
PKIX_ProcessingParams_GetRevocationCheckers;
PKIX_ProcessingParams_GetTargetCertConstraints;
PKIX_ProcessingParams_GetTrustAnchors;
PKIX_ProcessingParams_IsAnyPolicyInhibited;
PKIX_ProcessingParams_IsCRLRevocationCheckingEnabled;
PKIX_ProcessingParams_IsExplicitPolicyRequired;
PKIX_ProcessingParams_IsPolicyMappingInhibited;
PKIX_ProcessingParams_SetAnyPolicyInhibited;
PKIX_ProcessingParams_SetCertChainCheckers;
PKIX_ProcessingParams_SetCertStores;
PKIX_ProcessingParams_SetDate;
PKIX_ProcessingParams_SetExplicitPolicyRequired;
PKIX_ProcessingParams_SetHintCerts;
PKIX_ProcessingParams_SetInitialPolicies;
PKIX_ProcessingParams_SetPolicyMappingInhibited;
PKIX_ProcessingParams_SetPolicyQualifiersRejected;
PKIX_ProcessingParams_SetResourceLimits;
PKIX_ProcessingParams_SetRevocationCheckers;
PKIX_ProcessingParams_SetRevocationEnabled;
PKIX_ProcessingParams_SetTargetCertConstraints;
PKIX_RevocationChecker_Create;
PKIX_RevocationChecker_GetRevCallback;
PKIX_RevocationChecker_GetRevCheckerContext;
PKIX_Shutdown;
pkix_Throw;
PKIX_TrustAnchor_CreateWithCert;
PKIX_TrustAnchor_CreateWithNameKeyPair;
PKIX_TrustAnchor_GetCAName;
PKIX_TrustAnchor_GetCAPublicKey;
PKIX_TrustAnchor_GetNameConstraints;
PKIX_TrustAnchor_GetTrustedCert;
PKIX_ValidateChain;
PKIX_ValidateChain_NB;
PKIX_ValidateParams_Create;
PKIX_ValidateParams_GetCertChain;
PKIX_ValidateParams_GetProcessingParams;
pkix_ValidateResult_Create;
PKIX_ValidateResult_GetPolicyTree;
PKIX_ValidateResult_GetPublicKey;
PKIX_ValidateResult_GetTrustAnchor;
pkix_VerifyNode_AddToChain;
pkix_VerifyNode_Create;
PKIX_ResourceLimits_Create;
PKIX_ResourceLimits_GetMaxDepth;
PKIX_ResourceLimits_GetMaxFanout;
PKIX_ResourceLimits_GetMaxTime;
PKIX_ResourceLimits_SetMaxDepth;
PKIX_ResourceLimits_SetMaxFanout;
PKIX_ResourceLimits_SetMaxTime;
;+};
